# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws large bounding boxes with tags, over identified groups of bounding boxes.
"""

from typing import Any, Dict, List

import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.bbox import draw_bboxes, draw_tags
from peekingduck.pipeline.nodes.draw.utils.constants import TOMATO


class Node(AbstractNode):
    """Draws large bounding boxes with tags over multiple object bounding boxes
    which have been identified as belonging to the same group.

    The :term:`large_groups` data type from :mod:`dabble.check_large_groups`,
    and the ``groups`` key of the :term:`obj_attrs` data type from
    :mod:`dabble.group_nearby_objs`, are inputs for this node which identifies
    the different groups, and the group associated with each bounding box.

    For better understanding, refer to
    the :doc:`Group Size Checking use case </use_cases/group_size_checking>`.

    Inputs:
        |img_data|

        |bboxes_data|

        |obj_attrs_data|
        :mod:`draw.group_bbox_and_tag` requires the ``groups`` attribute from
        :mod:`dabble.group_nearby_objs`.

        |large_groups_data|

    Outputs:
        |none_output_data|

    Configs:
        tag (:obj:`str`): **default = "LARGE GROUP!"**. |br|
            The string message printed when a large group is detected.

    .. versionchanged:: 1.2.0
        :mod:`draw.group_bbox_and_tag` used to take in ``obj_tags``
        (:obj:`List[str]`) as an input data type, which has been deprecated and
        now subsumed under :term:`obj_attrs`. The same attribute is accessed by
        using the ``groups`` key of :term:`obj_attrs`.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws large bounding boxes over multiple object bounding boxes
        which have been identified as belonging to large groups.

        Args:
            inputs (dict): Dictionary with keys "img", "bboxes", "obj_attrs",
                "large_groups".

        Returns:
            outputs (dict): Dictionary with keys "none".
        """
        group_bboxes = self._get_group_bbox_coords(
            inputs["large_groups"], inputs["bboxes"], inputs["obj_attrs"]["groups"]
        )
        group_tags = self._get_group_tags(inputs["large_groups"], self.tag)

        # show labels set to False to reduce clutter on display
        draw_bboxes(inputs["img"], group_bboxes, [], False, TOMATO)
        draw_tags(inputs["img"], group_bboxes, group_tags, TOMATO)

        return {}

    def _get_config_types(self) -> Dict[str, Any]:
        """Returns dictionary mapping the node's config keys to respective types."""
        return {"tag": str}

    @staticmethod
    def _get_group_bbox_coords(
        large_groups: List[int], bboxes: List[np.ndarray], obj_groups: List[int]
    ) -> List[np.ndarray]:
        """For bboxes that belong to the same large group, get the coordinates of
        a large bbox that combines all these individual bboxes. Repeat for all large
        groups.
        """
        group_bboxes = []
        bboxes = np.array(bboxes)
        obj_groups = np.array(obj_groups)
        for group in large_groups:
            # filter relevant bboxes, select top-left and bot-right corners
            group_bbox = np.array([1.0, 1.0, 0.0, 0.0])
            selected_bboxes = bboxes[obj_groups == group]
            group_bbox[:2] = np.amin(selected_bboxes, axis=0)[:2]
            group_bbox[2:] = np.amax(selected_bboxes, axis=0)[2:]
            group_bboxes.append(group_bbox)

        return group_bboxes

    @staticmethod
    def _get_group_tags(large_groups: List[int], tag: str) -> List[str]:
        """Creates a list of tags to be used for the draw_tags function."""
        group_tags = [tag for _ in range(len(large_groups))]

        return group_tags



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility helper methods for PeekingDuck Viewer"""

from PIL import ImageTk, Image


def get_keyboard_char(char: str, keysym: str) -> str:
    """Get keyboard character

    Args:
        char (str): Tk keypress event character
        keysym (str): Tk keypress event key symbol

    Returns:
        str: keyboard character
    """
    res = char if char else keysym
    if keysym == "minus":
        res = "-"
    if keysym == "plus":
        res = "+"
    return res


def get_keyboard_modifier(state: int) -> str:
    """Get keyboard modifier keys: support ctrl, alt/option, shift

    Args:
        state (int): Tk keypress event key state

    Returns:
        str: detected modifier keys

    Technotes:
        Modifier = {
            0x1    : 'Shift'            ,
            0x2    : 'CapsLock'         ,
            0x4    : 'Control'          ,
            0x8    : 'Left-Alt'         ,   #Mod1 (meta / mac Left-Option)
            0x10   : 'NumLock'          ,   #Mod2? (alt?)
            0x20   : 'ScrollLock'       ,   #Mod3? (keypad?)
            0x40   : 'Mod4'             ,   #?
            0x80   : 'Right-Alt'        ,   #Mod5 (meta / mac Right-Option)
            0x100  : 'MouseButton1'     ,
            0x200  : 'MouseButton2'     ,
            0x400  : 'MouseButton3'     ,
            0x800  : 'Button4'          ,
            0x1000 : 'Button5'          ,
            0x20000: 'Alt'              ,   #not a typo
        }
    """
    ctrl = (state & 0x4) != 0
    alt = (state & 0x8) != 0 or (state & 0x80) != 0
    shift = (state & 0x1) != 0
    res = f"{'ctrl' if ctrl else ''}{'-alt' if alt else ''}{'-shift' if shift else ''}"
    return "" if len(res) == 0 else res[1:] if res[0] == "-" else res


def load_image(image_path: str, resize_pct: float = 1.0) -> ImageTk.PhotoImage:
    """Load and resize an image, 'coz plain vanilla Tkinter doesn't support JPG, PNG

    Args:
        resize_pct (float, optional): percentage to resize.
                                      Defaults to original size 1.0.

    Returns:
        ImageTk.PhotoImage: the loaded image
    """
    img = Image.open(image_path)
    if resize_pct != 1.0:
        width = int(resize_pct * img.size[0])
        height = int(resize_pct * img.size[1])
        resized_img = img.resize((width, height))
    else:
        resized_img = img
    the_img = ImageTk.PhotoImage(resized_img)
    return the_img



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Shows the outputs on your display.
"""

from typing import Any, Dict, Union

import cv2
import numpy as np
from peekingduck.pipeline.nodes.abstract_node import AbstractNode

MIN_DISPLAY_SIZE = 120


class Node(AbstractNode):
    """Streams the output on your display.

    Inputs:
        |img_data|

        |filename_data|

    Outputs:
        |pipeline_end_data|

    Configs:
        window_name (:obj:`str`): **default = "PeekingDuck"** |br|
            Name of the displayed window.
        window_size (:obj:`Dict[str, Union[bool, int]]`):
            **default = { do_resizing: False, width: 1280, height: 720 }** |br|
            Resizes the displayed window to the chosen width and weight, if
            ``do_resizing`` is set to ``true``. The size of the displayed
            window can also be adjusted by clicking and dragging.
        window_loc (:obj:`Dict[str, int]`): **default = { x: 0, y: 0 }** |br|
            X and Y coordinates of the top left corner of the displayed window,
            with reference from the top left corner of the screen, in pixels.

    .. note::

        **See Also:**

        :ref:`PeekingDuck Viewer<pkd_viewer>`: a GUI for running PeekingDuck pipelines.

        .. figure:: /assets/diagrams/viewer_cat_computer.png

        The PeekingDuck Viewer offers a GUI to view and analyze pipeline output.
        It has controls to re-play output video, scrub to a frame of interest,
        zoom video, and a playlist for managing multiple pipelines.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.previous_filename = ""
        pkd_viewer = config["pkd_viewer"] if config is not None else False
        if not pkd_viewer:
            cv2.namedWindow(self.window_name, cv2.WINDOW_NORMAL)
            cv2.moveWindow(self.window_name, self.window_loc["x"], self.window_loc["y"])

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Show the outputs on your display"""
        img = inputs["img"]
        self._set_window_size(inputs["filename"], img)
        cv2.imshow(self.window_name, img)
        if cv2.waitKey(1) & 0xFF == ord("q"):
            cv2.destroyWindow(self.window_name)
            return {"pipeline_end": True}

        return {"pipeline_end": False}

    def _set_window_size(self, current_filename: str, img: np.ndarray) -> None:
        """If `do_resizing` option is False, window size will be initialized to the image or
        video's frame default size for every new video or image.

        If `do_resizing` option is True, window size will be initialized to the config setting's
        width and height for every new video or image.

        The length of either sides of the display window will be clamped to a lower bound of
        `MIN_DISPLAY_SIZE`

        Args:
            current_filename (str): The filename from the `inputs` dictionary
            img (np.ndarray): The current image. The image will not be changed in this function.
        """
        if current_filename != self.previous_filename:
            # Initialize the window size for every new video
            if self.window_size["do_resizing"]:
                # Clamp the sides fo the window_size to have a minimum of MIN_DISPLAY_SIZE
                img_width = max(self.window_size["width"], MIN_DISPLAY_SIZE)
                img_height = max(self.window_size["height"], MIN_DISPLAY_SIZE)
            else:
                img_height, img_width, _ = img.shape
            cv2.resizeWindow(self.window_name, img_width, img_height)
            self.previous_filename = current_filename
        else:
            _, _, win_width, win_height = cv2.getWindowImageRect(self.window_name)
            if win_width < MIN_DISPLAY_SIZE or win_height < MIN_DISPLAY_SIZE:
                win_width = max(win_width, MIN_DISPLAY_SIZE)
                win_height = max(win_height, MIN_DISPLAY_SIZE)
                cv2.resizeWindow(self.window_name, win_width, win_height)

    def _get_config_types(self) -> Dict[str, Any]:
        """Returns dictionary mapping the node's config keys to respective types."""
        return {
            "window_name": str,
            "window_loc": Dict[str, int],
            "window_loc.x": int,
            "window_loc.y": int,
            "window_size": Dict[str, Union[bool, int]],
            "window_size.do_resizing": bool,
            "window_size.width": int,
            "window_size.height": int,
        }



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Implements the FrozenBatchNorm2d module."""

from typing import List
import torch
from torch import nn, Tensor


class FrozenBatchNorm2d(nn.Module):
    """
    BatchNorm2d where the batch statistics and the affine parameters
    are fixed
    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
    ):
        super().__init__()
        self.eps = eps
        self.weight: Tensor
        self.bias: Tensor
        self.running_var: Tensor
        self.running_mean: Tensor
        self.register_buffer("weight", torch.ones(num_features))
        self.register_buffer("bias", torch.zeros(num_features))
        self.register_buffer("running_mean", torch.zeros(num_features))
        self.register_buffer("running_var", torch.ones(num_features))

    def _load_from_state_dict(
        self,
        state_dict: dict,
        prefix: str,
        local_metadata: dict,
        strict: bool,
        missing_keys: List[str],
        unexpected_keys: List[str],
        error_msgs: List[str],
    ) -> None:
        # pylint: disable=too-many-arguments
        """Overrides _load_from_state_dict method from nn.Module parent class.
        Removes parameters with names prefix + "num_batches_tracked" (related to FrozenBatchNorm2d)
        from state_dict before loading it into the model
        """
        num_batches_tracked_key = prefix + "num_batches_tracked"
        if num_batches_tracked_key in state_dict:
            del state_dict[num_batches_tracked_key]

        super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward propagation for 2D Batch Normalization

        Args:
            inputs (Tensor): Batch input

        Returns:
            Tensor: Batch Normalized Tensor
        """
        # move reshapes to the beginning to make it user-friendly
        weight = self.weight.reshape(1, -1, 1, 1)
        bias = self.bias.reshape(1, -1, 1, 1)
        running_var = self.running_var.reshape(1, -1, 1, 1)
        running_mean = self.running_mean.reshape(1, -1, 1, 1)
        scale = weight * (running_var + self.eps).rsqrt()
        adjusted_bias = bias - running_mean * scale
        return inputs * scale + adjusted_bias

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.weight.shape[0]}, eps={self.eps})"



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Removes distortion from a wide-angle camera image.
"""

from pathlib import Path
from typing import Any, Dict

import cv2
import numpy as np
import yaml

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Undistorts an image by removing `radial and tangential distortion
    <https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html>`_.
    This may help to improve the performance of certain models.

    Before using this node for the first time, please follow the tutorial in
    :mod:`dabble.camera_calibration` to calculate the camera coefficients
    of the camera you are using, and ensure that the file_path that the
    coefficients are stored in is the same as the one specified in the configs.

    The images below show an example of an image before and after undistortion.
    Note that after undistortion, the shape of the image will change and the FOV
    will be reduced slightly.

    .. figure:: /assets/api/undistort.png
        :width: 80 %

        Before undistortion (left) and after undistortion (right)

    Inputs:
        |img_data|

    Outputs:
        |img_data|

    Configs:
        file_path (:obj:`str`):
            **default = "PeekingDuck/data/camera_calibration_coeffs.yml"**. |br|
            Path of the YML file containing calculated camera coefficients.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.file_path = Path(self.file_path)  # type: ignore
        if self.file_path.suffix != ".yml":
            raise ValueError("Filepath must have a '.yml' extension.")
        if not self.file_path.exists():
            raise FileNotFoundError(
                f"File {self.file_path} does not exist. Please run the camera calibration again "
                "with the dabble.camera_calibation node. You may refer to this tutorial: "
                "https://peekingduck.readthedocs.io/en/stable/nodes/dabble.camera_calibration.html"
            )

        yaml_file = yaml.safe_load(open(self.file_path))
        self.camera_matrix = np.array(yaml_file["camera_matrix"])
        self.distortion_coeffs = np.array(yaml_file["distortion_coeffs"])

        self.new_camera_matrix = None

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Undistorts an image by removing radial distortion

        Args:
            inputs (dict): Inputs dictionary with the key `img`.

        Returns:
            outputs (dict): Outputs dictionary with the key `img`.
        """
        img = inputs["img"]

        if self.new_camera_matrix is None:
            height, width = img.shape[:2]
            self.new_camera_matrix, self.roi = cv2.getOptimalNewCameraMatrix(
                self.camera_matrix,
                self.distortion_coeffs,
                (width, height),
                0,
                (width, height),
            )

        undistorted_img = cv2.undistort(
            img,
            self.camera_matrix,
            self.distortion_coeffs,
            None,
            self.new_camera_matrix,
        )

        x_pos, y_pos, img_w, img_h = self.roi
        undistorted_img = undistorted_img[y_pos : y_pos + img_h, x_pos : x_pos + img_w]

        return {"img": undistorted_img}

    def _get_config_types(self) -> Dict[str, Any]:
        """Returns dictionary mapping the node's config keys to respective types."""
        return {"file_path": str}



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Box operations for Region Proposal Network
Modifications include:
- Removed unused functions from original script
    - _box_inter_union
    - box_iou
- Removed tracing related code
"""

from typing import Tuple
import torch
from torch import Tensor
import torchvision
from torchvision.extension import _assert_has_ops


# pylint: disable=protected-access
def nms(boxes: Tensor, scores: Tensor, iou_threshold: float) -> Tensor:
    """
    Performs non-maximum suppression (NMS) on the boxes according
    to their intersection-over-union (IoU).

    NMS iteratively removes lower scoring boxes which have an
    IoU greater than iou_threshold with another (higher scoring)
    box.

    If multiple boxes have the exact same score and satisfy the IoU
    criterion with respect to a reference box, the selected box is
    not guaranteed to be the same between CPU and GPU. This is similar
    to the behavior of argsort in PyTorch when repeated values are present.

    Args:
        boxes (Tensor[N, 4])): boxes to perform NMS on. They
            are expected to be in ``(x1, y1, x2, y2)`` format with ``0 <= x1 < x2`` and
            ``0 <= y1 < y2``.
        scores (Tensor[N]): scores for each one of the boxes
        iou_threshold (float): discards all overlapping boxes with IoU > iou_threshold

    Returns:
        Tensor: int64 tensor with the indices of the elements that have been kept
        by NMS, sorted in decreasing order of scores
    """
    _assert_has_ops()
    return torch.ops.torchvision.nms(boxes, scores, iou_threshold)


def clip_boxes_to_image(boxes: Tensor, size: Tuple[int, int]) -> Tensor:
    """
    Clip boxes so that they lie inside an image of size `size`.

    Args:
        boxes (Tensor[N, 4]): boxes in ``(x1, y1, x2, y2)`` format
            with ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
        size (Tuple[height, width]): size of the image

    Returns:
        Tensor[N, 4]: clipped boxes
    """
    dim = boxes.dim()
    boxes_x = boxes[..., 0::2]
    boxes_y = boxes[..., 1::2]
    height, width = size

    boxes_x = boxes_x.clamp(min=0, max=width)
    boxes_y = boxes_y.clamp(min=0, max=height)

    clipped_boxes = torch.stack((boxes_x, boxes_y), dim=dim)
    return clipped_boxes.reshape(boxes.shape)


def remove_small_boxes(boxes: Tensor, min_size: float) -> Tensor:
    """
    Remove boxes which contains at least one side smaller than min_size.

    Args:
        boxes (Tensor[N, 4]): boxes in ``(x1, y1, x2, y2)`` format
            with ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
        min_size (float): minimum size

    Returns:
        Tensor[K]: indices of the boxes that have both sides
        larger than min_size
    """
    widths, heights = boxes[:, 2] - boxes[:, 0], boxes[:, 3] - boxes[:, 1]
    keep = (widths >= min_size) & (heights >= min_size)
    keep = torch.where(keep)[0]
    return keep


def batched_nms(
    boxes: Tensor,
    scores: Tensor,
    idxs: Tensor,
    iou_threshold: float,
) -> Tensor:
    """
    Performs non-maximum suppression in a batched fashion.

    Each index value correspond to a category, and NMS
    will not be applied between elements of different categories.

    Args:
        boxes (Tensor[N, 4]): boxes where NMS will be performed. They
            are expected to be in ``(x1, y1, x2, y2)`` format with ``0 <= x1 < x2`` and
            ``0 <= y1 < y2``.
        scores (Tensor[N]): scores for each one of the boxes
        idxs (Tensor[N]): indices of the categories for each one of the boxes.
        iou_threshold (float): discards all overlapping boxes with IoU > iou_threshold

    Returns:
        Tensor: int64 tensor with the indices of the elements that have been kept by NMS, sorted
        in decreasing order of scores
    """
    # Benchmarks that drove the following thresholds are at
    # https://github.com/pytorch/vision/issues/1311#issuecomment-781329339
    # Ideally for GPU we'd use a higher threshold
    if boxes.numel() > 4000 and not torchvision._is_tracing():
        return _batched_nms_vanilla(boxes, scores, idxs, iou_threshold)

    return _batched_nms_coordinate_trick(boxes, scores, idxs, iou_threshold)


@torch.jit._script_if_tracing
def _batched_nms_coordinate_trick(
    boxes: Tensor,
    scores: Tensor,
    idxs: Tensor,
    iou_threshold: float,
) -> Tensor:
    """strategy: in order to perform NMS independently per class,
    we add an offset to all the boxes. The offset is dependent
    only on the class idx, and is large enough so that boxes
    from different classes do not overlap"""
    if boxes.numel() == 0:
        return torch.empty((0,), dtype=torch.int64, device=boxes.device)
    max_coordinate = boxes.max()
    offsets = idxs.to(boxes) * (max_coordinate + torch.tensor(1).to(boxes))
    boxes_for_nms = boxes + offsets[:, None]
    keep = nms(boxes_for_nms, scores, iou_threshold)
    return keep


@torch.jit._script_if_tracing
def _batched_nms_vanilla(
    boxes: Tensor,
    scores: Tensor,
    idxs: Tensor,
    iou_threshold: float,
) -> Tensor:
    # Based on Detectron2 implementation, just manually call nms() on each class independently
    keep_mask = torch.zeros_like(scores, dtype=torch.bool)
    for class_id in torch.unique(idxs):
        curr_indices = torch.where(idxs == class_id)[0]
        curr_keep_indices = nms(
            boxes[curr_indices], scores[curr_indices], iou_threshold
        )
        keep_mask[curr_indices[curr_keep_indices]] = True
    keep_indices = torch.where(keep_mask)[0]
    return keep_indices[scores[keep_indices].sort(descending=True)[1]]


def box_area(boxes: Tensor) -> Tensor:
    """
    Computes the area of a set of bounding boxes, which are specified by their
    (x1, y1, x2, y2) coordinates.

    Args:
        boxes (Tensor[N, 4]): boxes for which the area will be computed. They
            are expected to be in (x1, y1, x2, y2) format with
            ``0 <= x1 < x2`` and ``0 <= y1 < y2``.

    Returns:
        Tensor[N]: the area for each box
    """
    boxes = _upcast(boxes)
    return (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])


def _upcast(input_tensor: Tensor) -> Tensor:
    """Protects from numerical overflows in multiplications by upcasting to the equivalent
    higher type"""
    if input_tensor.is_floating_point():
        return (
            input_tensor
            if input_tensor.dtype in (torch.float32, torch.float64)
            else input_tensor.float()
        )

    return (
        input_tensor
        if input_tensor.dtype in (torch.int32, torch.int64)
        else input_tensor.int()
    )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of bboxes and masks for mask_rcnn
"""

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import torch.backends as cudnn
from torch import Tensor
import torch.nn.functional as F
import torch

import numpy as np

from peekingduck.pipeline.nodes.model.yolact_edgev1.yolact_edge_files.model import (
    YolactEdge,
)
from peekingduck.pipeline.nodes.model.yolact_edgev1.yolact_edge_files.utils import (
    FastBaseTransform,
    crop,
)


class Detector:  # pylint: disable=too-many-instance-attributes
    """Detector class to handle detection of bboxes and masks for YolactEdge

    Attributes:
        logger (logging.Logger): Events logger.
        config (Dict[str, Any]): YolactEdge node configuration.
        model_dir (pathlib.Path): Path to directory of model weights files.
        device (torch.device): Represents the device on which the Tensor
            will be allocated.
        yolact_edge (YolactEdge): The YolactEdge model for performing inference.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: List[str],
        detect_ids: List[int],
        model_type: str,
        num_classes: int,
        model_file: Dict[str, str],
        input_size: int,
        max_num_detections: int,
        score_threshold: float,
        iou_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.device_is_cuda: bool = torch.cuda.is_available()
        self.device = torch.device("cuda" if self.device_is_cuda else "cpu")
        self.class_names = class_names
        self.detect_ids = detect_ids
        self.detect_ids_tensor = torch.tensor(
            detect_ids, dtype=torch.int64, device=self.device
        )
        self.model_type = model_type
        self.num_classes = num_classes
        self.model_path = model_dir / model_file[self.model_type]
        self.input_size = (input_size, input_size)
        self.max_num_detections = max_num_detections
        self.score_threshold = score_threshold
        self.iou_threshold = iou_threshold

        self.update_detect_ids(detect_ids)
        self.yolact_edge = self._create_yolact_edge_model()
        self.filtered_output: Dict[str, torch.Tensor] = dict()

    @torch.no_grad()
    def predict_instance_mask_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """YolactEdge masks and bboxes prediction function

        Args:
            image (np.ndarray): image in numpy array.

        Returns:
            bboxes (np.ndarray): array of detected bboxes
            labels (np.ndarray): array of labels
            scores (np.ndarray): array of scores
            masks (np.ndarray): array of detected masks
        """
        with torch.no_grad():
            if self.device_is_cuda:
                cudnn.benchmark = True
                cudnn.fastest = True
                cudnn.deterministic = True
                torch.set_default_tensor_type("torch.cuda.FloatTensor")
                frame = torch.from_numpy(image).cuda().float()
                torch.cuda.synchronize()
            else:
                torch.set_default_tensor_type("torch.FloatTensor")
                frame = torch.from_numpy(image).float()
        img_shape = image.shape[:2]
        model = self.yolact_edge

        preds = model(FastBaseTransform(self.input_size)(frame.unsqueeze(0)))[
            "pred_outs"
        ]
        labels, scores, boxes, masks = self._postprocess(preds[0], img_shape)

        return boxes, labels, scores, masks

    def update_detect_ids(self, ids: List[int]) -> None:
        """Updates list of selected object category IDs. When the list is
        empty, all available object category IDs are detected.

        Args:
            ids (List[int]): List of selected object category IDs
        """
        self.detect_ids = Tensor(ids).to(self.device)  # type: ignore

    def _create_yolact_edge_model(self) -> YolactEdge:
        """Creates YolactEdge model and loads its weights.
        Creates `detect_ids` as a `torch.Tensor`. Logs model configurations.

        Returns:
            (YolactEdge): YolactEdge model
        """
        self.logger.info(
            "YolactEdge model loaded with the following configs:\n\t"
            f"Model type: {self.model_type}\n\t"
            f"Input resolution: {self.input_size}\n\t"
            f"IDs being detected: {self.detect_ids.int().tolist()}\n\t"
            f"Score threshold: {self.score_threshold}\n\t"
            f"IOU threshold: {self.iou_threshold}"
        )

        return self._load_yolact_edge_weights()

    def _get_model(self) -> YolactEdge:
        """Constructs YolactEdge model based on parsed configuration.

        Returns:
            (YolactEdge): YolactEdge model.
        """
        return YolactEdge(
            self.model_type,
            self.input_size[0],
            self.iou_threshold,
            self.max_num_detections,
        )

    def _load_yolact_edge_weights(self) -> YolactEdge:
        """Loads YolactEdge model weights.

        Returns:
            (YolactEdge): YolactEdge model.

        Raises:
            ValueError: `model_path` does not exist.
        """
        if self.model_path.is_file():
            model = self._get_model()
            model.load_weights(self.model_path)
            model.eval()
            if self.device_is_cuda:
                model = model.cuda()
            return model

        raise ValueError(
            f"Model file does not exist. Please check that {self.model_path} exists"
        )

    def _postprocess(  # pylint: disable=too-many-locals
        self,
        network_output: Dict[str, Tensor],
        img_shape: Tuple[int, ...],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Postprocessing of detected bboxes and masks for YolactEdge

        Args:
            network_output (Dict[str, Tensor]): A dictionary from the first
                element of the YolactEdge output. The keys are "class", "box",
                "score", "mask", and "proto"
            img_shape (Tuple[int, int]): height and width of original image

        Returns:
            labels (ndarray): An array of human-friendly detection class names
            scores (ndarray): An array of confidence scores of the detections
            boxes (ndarray): An array of detection boxes of x1, y1, x2, y2 coords
            masks (ndarray): An array of masks in uint8
        """
        try:
            keep = network_output["score"] > self.score_threshold
            for k in network_output:
                if k != "proto":
                    network_output[k] = network_output[k][keep]

            detect_filter = torch.where(
                torch.isin(network_output["class"], self.detect_ids_tensor)
            )

            for output_key in network_output.keys():
                self.filtered_output[output_key] = network_output[output_key][
                    detect_filter
                ]

            classes = network_output["class"]
            box = network_output["box"]
            score = network_output["score"]
            mask = network_output["mask"]
            proto_data = network_output["proto"]

            mask = proto_data @ mask.t()
            mask = torch.sigmoid(mask)
            mask = crop(mask, box)
            mask = mask.permute(2, 0, 1).contiguous()
            mask = (
                F.interpolate(
                    mask.unsqueeze(0),
                    (img_shape[0], img_shape[1]),
                    mode="bilinear",
                    align_corners=False,
                )
                .squeeze(0)
                .gt_(0.5)
            )

            # Filters the detections to the IDs being detected as specified in the config
            labels = np.array([self.class_names[i] for i in classes[detect_filter]])
            boxes = np.array(box[detect_filter].cpu())
            boxes = np.clip(boxes, 0, 1)
            scores = np.array(score[detect_filter].cpu())
            masks = np.array(mask[detect_filter].cpu()).astype(np.uint8)

        except TypeError:
            return (
                np.empty((0)),
                np.empty((0), dtype=np.float32),
                np.empty((0, 4), dtype=np.float32),
                np.empty((0, 0, 0), dtype=np.uint8),
            )

        except RuntimeError:
            return (
                np.empty((0)),
                np.empty((0), dtype=np.float32),
                np.empty((0, 4), dtype=np.float32),
                np.empty((0, 0, 0), dtype=np.uint8),
            )

        return labels, scores, boxes, masks



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Generate a ResNet model for Mask-RCNN.

Modifications include:
- Removed redundant functions that are not used in Mask-RCNN.
    - load_state_dict_from_url
    - Other ResNet variants that are not used in Mask-RCNN.
    - _resnet is subsumed into resnet50 and resnet101.
    - _forward_impl is subsumed into forward method for the ResNet class
- Removed unused arguments.
    - pretrained option
    - progress option
"""

from typing import Type, Any, Callable, Union, List, Optional
import torch
from torch import Tensor, nn


__all__ = ["resnet50", "resnet101"]


def conv3x3(
    in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1
) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    """Basic Block for ResNet"""

    expansion: int = 1
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward propagation for BasicBlock of ResNet

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tensor: Output Tensor
        """
        identity = inputs

        out = self.conv1(inputs)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(inputs)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    """Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    while original implementation places the stride at the first 1x1 convolution(self.conv1)
    according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    This variant is also known as ResNet V1.5 and improves accuracy according to
    https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward propagation of the Bottleneck blocks of ResNet

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tensor: Output Tensor
        """
        identity = inputs

        out = self.conv1(inputs)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(inputs)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    """Implements ResNet backbone"""

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(
            3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(
            block, 128, layers[1], stride=2, dilate=replace_stride_with_dilation[0]
        )
        self.layer3 = self._make_layer(
            block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1]
        )
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2]
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fully_connected = nn.Linear(512 * block.expansion, num_classes)

        for mod in self.modules():
            if isinstance(mod, nn.Conv2d):
                nn.init.kaiming_normal_(mod.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(mod, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(mod.weight, 1)
                nn.init.constant_(mod.bias, 0)

        # Zero-initialize the last BN in each residual branch, so that the residual
        # branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for mod in self.modules():
                if isinstance(mod, Bottleneck):
                    nn.init.constant_(mod.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(mod, BasicBlock):
                    nn.init.constant_(mod.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
    ) -> nn.Sequential:
        """Method to make layers for ResNet"""
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward propagation of the main ResNet model

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tensor: Output
        """
        inputs = self.conv1(inputs)
        inputs = self.bn1(inputs)
        inputs = self.relu(inputs)
        inputs = self.maxpool(inputs)

        inputs = self.layer1(inputs)
        inputs = self.layer2(inputs)
        inputs = self.layer3(inputs)
        inputs = self.layer4(inputs)

        inputs = self.avgpool(inputs)
        inputs = torch.flatten(inputs, 1)
        return self.fully_connected(inputs)


def resnet50(**kwargs: Any) -> ResNet:
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(**kwargs: Any) -> ResNet:
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    return ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PeekingDuck CLI `nodes` command."""

import math
from pathlib import Path
from typing import Tuple

import click

from peekingduck.declarative_loader import PEEKINGDUCK_NODE_TYPES


@click.command()
@click.argument("type_name", required=False)
def nodes(type_name: str = None) -> None:
    """Lists available nodes in PeekingDuck. When no argument is given, all
    available nodes will be listed. When the node type is given as an argument,
    all available nodes in the specified node type will be listed.

    Args:
        type_name (str): input, augment, model, draw, dabble, or output.
    """
    config_dir = Path(__file__).resolve().parents[1] / "configs"

    if type_name is None:
        node_types = PEEKINGDUCK_NODE_TYPES
    else:
        node_types = [type_name]

    for node_type in node_types:
        click.secho("\nPeekingDuck has the following ", bold=True, nl=False)
        click.secho(f"{node_type} ", fg="red", bold=True, nl=False)
        click.secho("nodes:", bold=True)

        node_names = [path.stem for path in (config_dir / node_type).glob("*.yml")]
        idx_and_node_names = list(enumerate(node_names, start=1))

        max_length = _len_enumerate(max(idx_and_node_names, key=_len_enumerate))
        for idx, node_name in idx_and_node_names:
            url = _get_node_url(node_type, node_name)
            node_name_width = max_length + 1 - _num_digits(idx)

            click.secho(f"{idx}:", nl=False)
            click.secho(f"{node_name: <{node_name_width}}", bold=True, nl=False)
            click.secho("Info: ", fg="yellow", nl=False)
            click.secho(url)

    click.secho("\n")


def _get_node_url(node_type: str, node_name: str) -> str:
    """Constructs the URL to documentation of the specified node.

    Args:
        node_type (str): One of input, model, dabble, draw or output.
        node_name (str): Name of the node.

    Returns:
        (str): Full URL to the documentation of the specified node.
    """
    node_path = f"{node_type}.{node_name}"
    url_prefix = "https://peekingduck.readthedocs.io/en/stable/nodes/"
    url_postfix = ".html#module-"

    return f"{url_prefix}{node_path}{url_postfix}{node_path}"


def _len_enumerate(item: Tuple) -> int:
    """Calculates the string length of an enumerate item while accounting for
    the number of digits of the index.

    Args:
        item (Tuple): An item return by ``enumerate``, contains
            ``(index, element)`` where ``element`` is a ``str``.

    Returns:
        (int): Sum of length of ``element`` and the number of digits of its
            index.
    """
    return _num_digits(item[0]) + len(item[1])


def _num_digits(number: int) -> int:
    """The number of digits of the given number.

    Args:
        number (int): The given number. It is assumed ``number > 0``.

    Returns:
        (int): Number of digits in the given number.
    """
    return int(math.log10(number) + 1)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Region Proposal Network for Faster R-CNN

Modifications include:
- Removed training related code
- Removed training related arguments from RegionProposalNetwork class
    - fg_iou_thresh
    - bg_iou_thresh
    - batch_size_per_image
    - positive_fraction
- Removed tracing related code
"""

from typing import Dict, List, Tuple
import torch
from torch import nn, Tensor
from torch.nn import functional as F
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import (
    boxes as box_ops,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection import (
    image_list,
    detection_utils as det_utils,
)


def permute_and_flatten(
    layer: Tensor, num_instance: int, channels: int, height: int, width: int
) -> Tensor:
    """Permutes a feature output to be the same format as the labels"""
    layer = layer.view(num_instance, -1, channels, height, width)
    layer = layer.permute(0, 3, 4, 1, 2)
    layer = layer.reshape(num_instance, -1, channels)
    return layer


def concat_box_prediction_layers(
    box_cls: List[Tensor], box_regression: List[Tensor]
) -> Tuple[Tensor, Tensor]:
    """Permutes each feature level output to be the same format as labels and concatenates them on
    the first dimension"""
    box_cls_flattened = []
    box_regression_flattened = []
    # for each feature level, permute the outputs to make them be in the
    # same format as the labels. Note that the labels are computed for
    # all feature levels concatenated, so we keep the same representation
    # for the objectness and the box_regression
    for box_cls_per_level, box_regression_per_level in zip(box_cls, box_regression):
        num_instance, anchors_x_channels, height, width = box_cls_per_level.shape
        anchors_x_4 = box_regression_per_level.shape[1]
        anchors = anchors_x_4 // 4
        channels = anchors_x_channels // anchors
        box_cls_per_level = permute_and_flatten(
            box_cls_per_level, num_instance, channels, height, width
        )
        box_cls_flattened.append(box_cls_per_level)

        box_regression_per_level = permute_and_flatten(
            box_regression_per_level, num_instance, 4, height, width
        )
        box_regression_flattened.append(box_regression_per_level)
    # concatenate on the first dimension (representing the feature levels), to
    # take into account the way the labels were generated (with all feature maps
    # being concatenated as well)
    box_cls_out = torch.cat(box_cls_flattened, dim=1).flatten(0, -2)
    box_regression_out = torch.cat(box_regression_flattened, dim=1).reshape(-1, 4)
    return box_cls_out, box_regression_out


class RPNHead(nn.Module):
    """
    Adds a simple RPN Head with classification and regression heads

    Args:
        in_channels (int): number of channels of the input feature
        num_anchors (int): number of anchors to be predicted
    """

    def __init__(self, in_channels: int, num_anchors: int):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels, in_channels, kernel_size=3, stride=1, padding=1
        )
        self.cls_logits = nn.Conv2d(in_channels, num_anchors, kernel_size=1, stride=1)
        self.bbox_pred = nn.Conv2d(
            in_channels, num_anchors * 4, kernel_size=1, stride=1
        )

        for layer in self.children():
            nn.init.normal_(layer.weight, std=0.01)  # type: ignore[arg-type]
            nn.init.constant_(layer.bias, 0)  # type: ignore[arg-type]

    def forward(self, inputs: List[Tensor]) -> Tuple[List[Tensor], List[Tensor]]:
        """Forward propagation for computing the objectness and bboxes' offsets from the anchor
        boxes regressions obtained from RPN

        Args:
            inputs (List[Tensor]): Input Tensor. length of inputs is the number of feature levels

        Returns:
            Tuple[List[Tensor], List[Tensor]]:
                - objectness logits. Length of list corresponds to the number of feature levels.
                    Each tensor element in the list has a shape of:
                    [
                        batch_size,
                        num_of_anchor_boxes_per_feature_level_per_location,
                        feature_height,
                        feature_width
                    ]
                - Bbox deltas regressions. Length of list corresponds to the number of feature
                    levels. Each tensor element in the list has a shape of:
                    [
                        batch_size,
                        num_of_anchor_boxes_per_feature_level_per_location x 4 corners of bbox,
                        feature_height,
                        feature_width
                    ]
        """
        logits = []
        bbox_reg = []
        for feature in inputs:
            transformed_feature = F.relu(self.conv(feature))
            logits.append(self.cls_logits(transformed_feature))
            bbox_reg.append(self.bbox_pred(transformed_feature))
        return logits, bbox_reg


class RegionProposalNetwork(nn.Module):
    """
    Implements Region Proposal Network (RPN).

    Args:
        anchor_generator (AnchorGenerator): module that generates the anchors for a set of feature
            maps.
        head (nn.Module): module that computes the objectness and regression deltas
        pre_nms_top_n (Dict[str, int]): number of proposals to keep before applying NMS. It should
            contain two fields: training and testing, to allow for different values depending
            on training or evaluation
        post_nms_top_n (Dict[str, int]): number of proposals to keep after applying NMS. It should
            contain two fields: training and testing, to allow for different values depending
            on training or evaluation
        nms_thresh (float): NMS threshold used for postprocessing the RPN proposals
        score_thresh (float): Score threshold for filtering low scoring proposals
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(
        self,
        anchor_generator: nn.Module,
        head: nn.Module,
        pre_nms_top_n: Dict[str, int],
        post_nms_top_n: Dict[str, int],
        nms_thresh: float,
        score_thresh: float = 0.0,
    ):
        super().__init__()
        self.anchor_generator = anchor_generator
        self.head = head
        self.box_coder = det_utils.BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

        # used during testing
        self._pre_nms_top_n = pre_nms_top_n
        self._post_nms_top_n = post_nms_top_n
        self.nms_thresh = nms_thresh
        self.score_thresh = score_thresh
        self.min_size = 1e-3

    def pre_nms_top_n(self) -> int:
        """Returns the number of proposals to keep before applying NMS"""
        return self._pre_nms_top_n["testing"]

    def post_nms_top_n(self) -> int:
        """Returns the number of proposals to keep after applying NMS"""
        return self._post_nms_top_n["testing"]

    def _get_top_n_idx(
        self, objectness: Tensor, num_anchors_per_level: List[int]
    ) -> Tensor:
        """Get the top n number of proposals before applying NMS"""
        results = []
        offset = 0
        for objectness_per_level in objectness.split(num_anchors_per_level, 1):
            num_anchors = objectness_per_level.shape[1]
            pre_nms_top_n = min(self.pre_nms_top_n(), num_anchors)
            _, top_n_idx = objectness_per_level.topk(pre_nms_top_n, dim=1)
            results.append(top_n_idx + offset)
            offset += num_anchors
        return torch.cat(results, dim=1)

    def filter_proposals(
        self,
        proposals: Tensor,
        objectness: Tensor,
        image_shapes: List[Tuple[int, int]],
        num_anchors_per_level: List[int],
    ) -> Tuple[List[Tensor], List[Tensor]]:
        # pylint: disable=too-many-locals
        """Filters proposals through objectness thresholds, NMS, minimum size"""
        num_images = proposals.shape[0]
        device = proposals.device
        # do not backprop through objectness
        objectness = objectness.detach()
        objectness = objectness.reshape(num_images, -1)

        levels = [
            torch.full((n,), idx, dtype=torch.int64, device=device)
            for idx, n in enumerate(num_anchors_per_level)
        ]
        levels_tensor = torch.cat(levels, 0)
        levels_tensor = levels_tensor.reshape(1, -1).expand_as(objectness)

        # select top_n boxes independently per level before applying nms
        top_n_idx = self._get_top_n_idx(objectness, num_anchors_per_level)

        image_range = torch.arange(num_images, device=device)
        batch_idx = image_range[:, None]

        objectness = objectness[batch_idx, top_n_idx]
        levels_tensor = levels_tensor[batch_idx, top_n_idx]
        proposals = proposals[batch_idx, top_n_idx]

        objectness_prob = torch.sigmoid(objectness)

        final_boxes = []
        final_scores = []
        for boxes, scores, lvl, img_shape in zip(
            proposals, objectness_prob, levels, image_shapes
        ):
            boxes = box_ops.clip_boxes_to_image(boxes, img_shape)

            # remove small boxes
            keep = box_ops.remove_small_boxes(boxes, self.min_size)
            boxes, scores, lvl = boxes[keep], scores[keep], lvl[keep]

            # remove low scoring boxes
            # use >= for Backwards compatibility
            keep = torch.where(scores >= self.score_thresh)[0]
            boxes, scores, lvl = boxes[keep], scores[keep], lvl[keep]

            # non-maximum suppression, independently done per level
            keep = box_ops.batched_nms(boxes, scores, lvl, self.nms_thresh)

            # keep only topk scoring predictions
            keep = keep[: self.post_nms_top_n()]
            boxes, scores = boxes[keep], scores[keep]

            final_boxes.append(boxes)
            final_scores.append(scores)
        return final_boxes, final_scores

    def forward(
        self,
        images: image_list.ImageList,
        features: Dict[str, Tensor],
    ) -> List[Tensor]:
        """
        Args:
            images (ImageList): images for which we want to compute the predictions
            features (OrderedDict[Tensor]): features computed from the images that are used for
                computing the predictions. Each tensor in the list correspond to different feature
                levels

        Returns:
            boxes (List[Tensor]): the predicted boxes from the RPN, one Tensor per image.
        """
        # RPN uses all feature maps that are available
        features_list = list(features.values())
        objectness, pred_bbox_deltas = self.head(features_list)
        anchors = self.anchor_generator(images, features_list)

        num_images = len(anchors)
        num_anchors_per_level_shape_tensors = [
            obj_per_lvl[0].shape for obj_per_lvl in objectness
        ]
        num_anchors_per_level = [
            shape[0] * shape[1] * shape[2]
            for shape in num_anchors_per_level_shape_tensors
        ]
        objectness, pred_bbox_deltas = concat_box_prediction_layers(
            objectness, pred_bbox_deltas
        )
        # apply pred_bbox_deltas to anchors to obtain the decoded proposals
        # note that we detach the deltas because Faster R-CNN do not backprop through
        # the proposals
        proposals = self.box_coder.decode(pred_bbox_deltas.detach(), anchors)
        proposals = proposals.view(num_images, -1, 4)
        boxes, _ = self.filter_proposals(
            proposals, objectness, images.image_sizes, num_anchors_per_level
        )

        return boxes



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Anchor Generator class for Faster-RCNN model

Modifications include:
- Removed conversion for sizes and aspect_ratios in __init__ of AnchorGenerator
"""

from typing import List, Tuple, Union
import torch
from torch import nn, Size, Tensor  # pylint: disable=no-name-in-module
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection.image_list import (
    ImageList,
)


class AnchorGenerator(nn.Module):
    """
    Module that generates anchors for a set of feature maps and
    image sizes.

    The module support computing anchors at multiple sizes and aspect ratios
    per feature map. This module assumes aspect ratio = height / width for
    each anchor.

    sizes and aspect_ratios should have the same number of elements, and it should
    correspond to the number of feature maps.

    sizes[i] and aspect_ratios[i] can have an arbitrary number of elements,
    and AnchorGenerator will output a set of sizes[i] * aspect_ratios[i] anchors
    per spatial location for feature map i.

    Args:
        sizes (Tuple[Tuple[int, ...], ...]): The sizes of the generated anchor boxes
        aspect_ratios (Tuple[Tuple[float, ...], ...]): The aspect ratios of the generated anchor
            boxes
    """

    def __init__(
        self,
        sizes: Tuple[Tuple[int, ...], ...] = ((128, 256, 512),),
        aspect_ratios: Tuple[Tuple[float, ...], ...] = ((0.5, 1.0, 2.0),),
    ):
        super().__init__()

        assert isinstance(sizes[0], (list, tuple))
        assert isinstance(aspect_ratios[0], (list, tuple))
        assert len(sizes) == len(aspect_ratios)

        self.sizes = sizes
        self.aspect_ratios = aspect_ratios
        self.cell_anchors = [
            self.generate_anchors(size, aspect_ratio)
            for size, aspect_ratio in zip(sizes, aspect_ratios)
        ]

    @staticmethod
    def generate_anchors(
        scales: Union[List[int], Tuple[int, ...]],
        aspect_ratios: Union[List[float], Tuple[float, ...]],
        dtype: torch.dtype = torch.float32,
        device: torch.device = torch.device("cpu"),
    ) -> Tensor:
        """Method to generate anchors"""
        scales_tensor = torch.as_tensor(scales, dtype=dtype, device=device)
        aspect_ratios_tensor = torch.as_tensor(
            aspect_ratios, dtype=dtype, device=device
        )
        height_ratios = torch.sqrt(aspect_ratios_tensor)
        width_ratios = 1 / height_ratios

        width_scales = (width_ratios[:, None] * scales_tensor[None, :]).view(-1)
        height_scales = (height_ratios[:, None] * scales_tensor[None, :]).view(-1)

        base_anchors = (
            torch.stack(
                [-width_scales, -height_scales, width_scales, height_scales], dim=1
            )
            / 2
        )
        return base_anchors.round()

    def set_cell_anchors(self, dtype: torch.dtype, device: torch.device) -> None:
        """Set anchors to target device"""
        self.cell_anchors = [
            cell_anchor.to(dtype=dtype, device=device)
            for cell_anchor in self.cell_anchors
        ]

    def num_anchors_per_location(self) -> List[int]:
        """Generate a list of number of anchors per location"""
        return [len(s) * len(a) for s, a in zip(self.sizes, self.aspect_ratios)]

    def grid_anchors(
        self, grid_sizes: List[Size], strides: List[List[Tensor]]
    ) -> List[Tensor]:
        # pylint: disable=too-many-locals
        """For every combination of (a, (g, s), i) in (self.cell_anchors,
        zip(grid_sizes, strides), 0:2), output g[i] anchors that are s[i] distance
        apart in direction i, with the same dimensions as a.
        """
        anchors = []
        cell_anchors = self.cell_anchors
        assert cell_anchors is not None

        if not len(grid_sizes) == len(strides) == len(cell_anchors):
            raise ValueError(
                "Anchors should be Tuple[Tuple[int]] because each feature "
                "map could potentially have different sizes and aspect ratios. "
                "There needs to be a match between the number of "
                "feature maps passed and the number of sizes / aspect ratios specified."
            )

        for size, stride, base_anchors in zip(grid_sizes, strides, cell_anchors):
            grid_height, grid_width = size
            stride_height, stride_width = stride
            device = base_anchors.device

            # For output anchor, compute [x_center, y_center, x_center, y_center]
            shifts_x = (
                torch.arange(0, grid_width, dtype=torch.int32, device=device)
                * stride_width
            )
            shifts_y = (
                torch.arange(0, grid_height, dtype=torch.int32, device=device)
                * stride_height
            )
            shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x, indexing="ij")
            shift_x = shift_x.reshape(-1)
            shift_y = shift_y.reshape(-1)
            shifts = torch.stack((shift_x, shift_y, shift_x, shift_y), dim=1)

            # For every (base anchor, output anchor) pair,
            # offset each zero-centered base anchor by the center of the output anchor.
            anchors.append(
                (shifts.view(-1, 1, 4) + base_anchors.view(1, -1, 4)).reshape(-1, 4)
            )

        return anchors

    def forward(
        self, image_list: ImageList, feature_maps: List[Tensor]
    ) -> List[Tensor]:
        """Generates anchors for a set of feature maps and image sizes.

        Args:
            image_list (ImageList):  An object that holds a list of padded images with its original
                image sizes
            feature_maps (List[Tensor]): List of feature maps for generating the anchors. Feature
                maps have the length equal to the number of feature levels, and the tensor of each
                element is in the shape of [batch_size, 256, feature_height, feature_width]. The
                number of channels 256 is based on the paper
                "Feature Pyramid Network for Object Detection" <https://arxiv.org/abs/1612.03144>
                under the sub-section "Top-down pathway and lateral connections", where the number
                of channels of the features from the top-down pathway is fixed to d=256.

        Returns:
            List[Tensor]: List of generated anchors. The number of elements in the list corresponds
                to the batch size and in each element, there are 4 columns, each corresponds to the
                coordinates of the anchor boxes [x0, y0, x1, y1].
        """
        grid_sizes = [feature_map.shape[-2:] for feature_map in feature_maps]
        image_size = image_list.tensors.shape[-2:]
        dtype, device = feature_maps[0].dtype, feature_maps[0].device
        strides = [
            [
                torch.tensor(image_size[0] // g[0], dtype=torch.int64, device=device),
                torch.tensor(image_size[1] // g[1], dtype=torch.int64, device=device),
            ]
            for g in grid_sizes
        ]
        self.set_cell_anchors(dtype, device)
        anchors_over_all_feature_maps = self.grid_anchors(grid_sizes, strides)
        anchors: List[List[Tensor]] = []
        for _ in range(len(image_list.image_sizes)):
            anchors.append(anchors_over_all_feature_maps)
        anchors_out = [torch.cat(anchors_per_image) for anchors_per_image in anchors]
        return anchors_out



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PeekingDuck CLI `create-node` command."""

import logging
from pathlib import Path

import click

from peekingduck.commands.create_node_helper import (
    create_config_and_script_files,
    ensure_relative_path,
    ensure_valid_name,
    ensure_valid_name_partial,
    ensure_valid_type,
    ensure_valid_type_partial,
    get_config_and_script_paths,
    verify_option,
)
from peekingduck.declarative_loader import PEEKINGDUCK_NODE_TYPES, DeclarativeLoader

logger = logging.getLogger("peekingduck.cli")  # pylint: disable=invalid-name


@click.command()
@click.option(
    "--node_subdir",
    help=(
        "Path to the custom nodes directory, relative to the directory where "
        "the command is invoked."
    ),
    required=False,
)
@click.option(
    "--node_type",
    help=(
        "Node type, only accepts values from existing node types defined in "
        f"{PEEKINGDUCK_NODE_TYPES}."
    ),
    required=False,
)
@click.option(
    "--node_name",
    help=(
        "\b\nName of new custom node. The name cannot be a duplicate of an \n"
        "existing custom node. The name has the following requirements:\n"
        "- Minimum 2 characters\n"
        "- Can only contain alphanumeric characters, dashes and underscores \n"
        "  /[[a-zA-Z0-9_\\-]/\n"
        "- Must start with an alphabet\n"
        "- Must end with an alphanumeric character"
    ),
    required=False,
)
@click.option(
    "--config_path",
    help=(
        "List of nodes to parse. Will automatically create the config and script file "
        "for any custom nodes."
    ),
    required=False,
)
def create_node(
    node_subdir: str = None,
    node_type: str = None,
    node_name: str = None,
    config_path: str = None,
) -> None:
    """Automates the creation of a new custom node.

    If the options `node_subdir`, `node_type`, or `node_name` are not
    specified, users will be prompted them for the values while performing
    checks to ensure value validity.
    """
    project_dir = Path.cwd()
    node_type_choices = click.Choice(PEEKINGDUCK_NODE_TYPES)

    if config_path is not None and any(
        arg is not None for arg in (node_subdir, node_type, node_name)
    ):
        raise ValueError(
            "--config_path cannot be use with --node_subdir, --node_type, or "
            "--node_name!"
        )
    if config_path is not None:
        _create_nodes_from_config_file(config_path, project_dir, node_type_choices)
    else:
        click.secho("Creating new custom node...")
        node_subdir = verify_option(node_subdir, value_proc=ensure_relative_path)
        if node_subdir is None:
            node_subdir = click.prompt(
                f"Enter node directory relative to {project_dir}/src",
                default="custom_nodes",
                value_proc=ensure_relative_path,
            )
        node_dir = project_dir / "src" / node_subdir

        node_type = verify_option(
            node_type, value_proc=ensure_valid_type_partial(node_type_choices)
        )
        if node_type is None:
            node_type = click.prompt(
                f"Select node type ({', '.join(node_type_choices.choices)})",
                value_proc=ensure_valid_type_partial(node_type_choices),
            )

        node_name = verify_option(
            node_name, value_proc=ensure_valid_name_partial(node_dir, node_type)
        )
        if node_name is None:
            node_name = click.prompt(
                "Enter node name",
                default="my_custom_node",
                value_proc=ensure_valid_name_partial(node_dir, node_type),
            )

        created_paths = get_config_and_script_paths(
            node_dir, ("configs", node_type), node_type, node_name
        )
        click.echo(f"\nNode directory:\t{node_dir}")
        click.echo(f"Node type:\t{node_type}")
        click.echo(f"Node name:\t{node_name}")
        click.echo("\nCreating the following files:")
        click.echo(f"\tConfig file: {created_paths['config']}")
        click.echo(f"\tScript file: {created_paths['script']}")

        proceed = click.confirm("Proceed?", default=True)
        if proceed:
            create_config_and_script_files(created_paths)
            click.echo("Created node!")
        else:
            click.echo("Aborted!")


def _create_nodes_from_config_file(
    config_path: str, project_dir: Path, node_type_choices: click.Choice
) -> None:
    """Creates custom nodes declared in the pipeline config file."""
    pipeline_path = Path(config_path)
    if not pipeline_path.is_absolute():
        pipeline_path = (project_dir / pipeline_path).resolve()
    if not pipeline_path.exists():
        raise FileNotFoundError(
            f"Config file '{config_path}' is not found at {pipeline_path}!"
        )
    logger.info(f"Creating custom nodes declared in {pipeline_path}.")
    # Load run config with DeclarativeLoader to ensure consistency
    loader = DeclarativeLoader(pipeline_path, "None", "src")
    try:
        node_subdir = project_dir / "src" / loader.custom_nodes_dir
    except AttributeError as custom_nodes_no_exist:
        raise ValueError(
            f"Config file '{config_path}' does not contain custom nodes!"
        ) from custom_nodes_no_exist

    for node_str, _ in loader.node_list:
        try:
            node_subdir, node_type, node_name = node_str.split(".")
        except ValueError:
            continue
        try:
            # Check node string formatting
            ensure_relative_path(node_subdir)
            node_dir = project_dir / "src" / node_subdir
            ensure_valid_type(node_type_choices, node_type)
            ensure_valid_name(node_dir, node_type, node_name)
        except click.exceptions.UsageError as err:
            logger.warning(
                f"{node_str} contains invalid formatting: '{err.message}'. "
                "Skipping..."
            )
            continue
        created_paths = get_config_and_script_paths(
            node_dir, ("configs", node_type), node_type, node_name
        )
        logger.info(
            f"Creating files for {node_str}:\n\t"
            f"Config file: {created_paths['config']}\n\t"
            f"Script file: {created_paths['script']}"
        )
        create_config_and_script_files(created_paths)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
functions for drawing legend related UI components
"""

from typing import Any, Dict, List, Tuple, Union

import cv2
import numpy as np
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA

from peekingduck.pipeline.nodes.draw.utils.constants import (
    BLACK,
    FILLED,
    PRIMARY_PALETTE,
    PRIMARY_PALETTE_LENGTH,
    WHITE,
)
from peekingduck.pipeline.nodes.draw.utils.general import get_image_size

ZONE_COUNTS_HEADING = "-ZONE COUNTS-"
LEGEND_LEFT_X = 15


class Legend:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """Legend class that uses available info to draw legend box on frame"""

    def __init__(
        self,
        items: List[str],
        position: str,
        box_opacity: float,
        font: Dict[str, Union[float, int]],
    ) -> None:
        self.items = items  # list of items to be drawn in legend box
        self.position = position
        self.box_opacity = box_opacity
        self.font_size = font["size"]
        self.font_thickness = font["thickness"]

        self.frame = None
        self.legend_starting_y = 0
        self.legend_width = 0
        self.legend_height = 0
        self.item_height = self._get_text_size("")[1]
        self.item_padding = self.item_height // 2

    def draw(self, inputs: Dict[str, Any]) -> None:
        """Draw legends onto image

        Args:
            inputs (dict): dictionary of all available inputs for drawing the legend
        """
        self.frame = inputs["img"]

        # legend box has to be drawn first to be "behind" text
        self._update_legend_size(inputs)
        self._set_legend_starting_y()
        self._draw_legend_box(self.frame)
        y_pos = self.legend_starting_y + self.item_height

        for item in self.items:
            if item == "zone_count":
                self._draw_zone_count(self.frame, y_pos, inputs[item])
            else:
                self._draw_item_info(self.frame, y_pos, item, inputs[item])
            y_pos += self.item_height + self.item_padding

    def _update_legend_size(self, inputs: Dict[str, Any]) -> None:
        """Update the width and height of the legend box"""
        self.legend_height = self._get_legend_height(inputs)
        self.legend_width = max(self.legend_width, self._get_legend_width(inputs))

    def _get_legend_height(self, inputs: Dict[str, Any]) -> int:
        """Get height of legend box needed to contain all items drawn"""
        num_items = len(self.items)
        if "zone_count" in self.items:
            # increase the number of items according to number of zones
            num_items += len(inputs["zone_count"])
        return (self.item_height + self.item_padding) * num_items

    def _get_legend_width(self, inputs: Dict[str, Any]) -> int:
        """Get width of legend box needed to contain all items drawn"""
        max_width = 0
        for item in self.items:
            if item != "zone_count":
                max_width = max(max_width, self._get_item_width(item, inputs[item]))
            else:
                max_width = self._get_item_width(ZONE_COUNTS_HEADING, "")
                for i, count in enumerate(inputs[item]):
                    max_width = max(
                        max_width,
                        self._get_item_width(f"ZONE-{i+1}", count)
                        + self.item_padding
                        + self.item_height,
                    )

        return max_width + 2 * self.item_padding

    def _get_item_width(
        self,
        item_name: str,
        item_info: Union[int, float, str],
    ) -> int:
        """Get width of the text to be drawn. If item info is
        of float type, it will be displayed in 2 decimal places.

        Args:
            item_name (str): name of the legend item
            item_info: Union[int, float, str]: info contained by the legend item
        """
        if not isinstance(item_info, (int, float, str)):
            raise TypeError(
                f"With the exception of the 'zone_count' data type, "
                f"the draw.legend node only draws values that are of type 'int', 'float' or 'str' "
                f"within the legend box. The value: {item_info} from the data type: {item_name} "
                f"is of type: {type(item_info)} and is unable to be drawn."
            )

        if isinstance(item_info, float):
            text = f"{item_name.upper()}: {item_info:.2f}"
        else:
            text = f"{item_name.upper()}: {str(item_info)}"

        return self._get_text_size(text)[0]

    def _set_legend_starting_y(self) -> None:
        assert self.legend_height != 0
        if self.position == "top":
            self.legend_starting_y = self.item_padding
        else:
            _, image_height = get_image_size(self.frame)
            self.legend_starting_y = (
                image_height - self.item_padding - self.legend_height
            )

    def _draw_legend_box(self, frame: np.ndarray) -> None:
        """draw pts of selected object onto frame

        Args:
            frame (np.array): image of current frame
        """
        assert self.legend_height is not None
        overlay = frame.copy()
        cv2.rectangle(
            overlay,
            (LEGEND_LEFT_X, self.legend_starting_y - self.item_padding),
            (
                LEGEND_LEFT_X + self.legend_width,
                self.legend_starting_y + self.legend_height,
            ),
            BLACK,
            FILLED,
        )
        # apply the overlay
        cv2.addWeighted(
            overlay, self.box_opacity, frame, 1 - self.box_opacity, 0, frame
        )

    def _draw_item_info(
        self,
        frame: np.ndarray,
        y_pos: int,
        item_name: str,
        item_info: Union[int, float, str],
    ) -> None:
        """Draw item name followed by item info onto frame. If item info is
        of float type, it will be displayed in 2 decimal places.

        Args:
            frame (np.array): image of current frame
            y_pos (int): y_position to draw the count text
            item_name (str): name of the legend item
            item_info: Union[int, float, str]: info contained by the legend item
        """
        if not isinstance(item_info, (int, float, str)):
            raise TypeError(
                f"With the exception of the 'zone_count' data type, "
                f"the draw.legend node only draws values that are of type 'int', 'float' or 'str' "
                f"within the legend box. The value: {item_info} from the data type: {item_name} "
                f"is of type: {type(item_info)} and is unable to be drawn."
            )

        if isinstance(item_info, float):
            text = f"{item_name.upper()}: {item_info:.2f}"
        else:
            text = f"{item_name.upper()}: {str(item_info)}"
        self._put_text(frame, text, (LEGEND_LEFT_X + self.item_padding, y_pos))

    def _draw_zone_count(
        self, frame: np.ndarray, y_pos: int, counts: List[int]
    ) -> None:
        """Draw zone counts of all zones onto frame image

        Args:
            frame (np.array): image of current frame
            y_pos (int): y position to draw the count info text
            counts (list): list of zone counts
        """
        self._put_text(
            frame, ZONE_COUNTS_HEADING, (LEGEND_LEFT_X + self.item_padding, y_pos)
        )
        for i, count in enumerate(counts):
            y_pos += self.item_height + self.item_padding
            cv2.rectangle(
                frame,
                (LEGEND_LEFT_X + self.item_padding, y_pos),
                (
                    LEGEND_LEFT_X + self.item_padding + self.item_height,
                    y_pos - self.item_height,
                ),
                PRIMARY_PALETTE[(i + 1) % PRIMARY_PALETTE_LENGTH],
                FILLED,
            )
            text = f" ZONE-{i+1}: {count}"
            self._put_text(
                frame,
                text,
                (LEGEND_LEFT_X + self.item_padding + self.item_height, y_pos),
            )

    def _get_text_size(self, text: str) -> Tuple[int, int]:
        """Wrapper around cv2.getTextSize method to reduce number of arguments in calls"""
        return cv2.getTextSize(
            text,
            FONT_HERSHEY_SIMPLEX,
            self.font_size,
            self.font_thickness,
        )[0]

    def _put_text(self, frame: np.ndarray, text: str, pos: Tuple[int, int]) -> None:
        """Wrapper around cv2.putText method to reduce number of arguments in calls"""
        cv2.putText(
            frame,
            text,
            pos,
            FONT_HERSHEY_SIMPLEX,
            self.font_size,
            WHITE,
            self.font_thickness,
            LINE_AA,
        )



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Multi-scale RoI Align class to handle RoIAlign for Mask-RCNN
Modifications include:
- Removed tracing related codes
- Renamed initLevelMapper function to init_level_mapper
"""

from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
import torch
from torch import nn, Tensor
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import (
    roi_align,
    boxes as box_ops,
)


def init_level_mapper(
    k_min: int,
    k_max: int,
    canonical_scale: int = 224,
    canonical_level: int = 4,
    eps: float = 1e-6,
) -> Callable:
    """Initialize the LevelMapper object"""
    return LevelMapper(k_min, k_max, canonical_scale, canonical_level, eps)


class LevelMapper:
    """Determine which FPN level each RoI in a set of RoIs should map to based
    on the heuristic in the FPN paper.

    Args:
        k_min (int): Minimum target level to be mapped to.
        k_max (int): Maximum target level to be mapped to.
        canonical_scale (int): The scale of the input image fed to the backbone.
        canonical_level (int): the target level of the pyramid that a region of interest with
            the size of w x h = canonical_scale x canonical_scale should be mapped into.
        eps (float): A very small number added for stability during calculation
    """

    # pylint: disable=too-many-arguments,too-few-public-methods
    def __init__(
        self,
        k_min: int,
        k_max: int,
        canonical_scale: int = 224,
        canonical_level: int = 4,
        eps: float = 1e-6,
    ):
        self.k_min = k_min
        self.k_max = k_max
        self.canon_scale = canonical_scale
        self.lvl0 = canonical_level
        self.eps = eps

    def __call__(self, boxlists: List[Tensor]) -> Tensor:
        """
        Args:
            boxlists (list[BoxList])
        """
        # Compute level ids
        scales = torch.sqrt(
            torch.cat([box_ops.box_area(boxlist) for boxlist in boxlists])
        )

        # Eqn.(1) in FPN paper
        target_lvls = torch.floor(
            self.lvl0
            + torch.log2(scales / self.canon_scale)
            + torch.tensor(self.eps, dtype=scales.dtype)
        )
        target_lvls = torch.clamp(target_lvls, min=self.k_min, max=self.k_max)
        return (target_lvls.to(torch.int64) - self.k_min).to(torch.int64)


class MultiScaleRoIAlign(nn.Module):
    """
    Multi-scale RoIAlign pooling, which is useful for detection with or without FPN.

    It infers the scale of the pooling via the heuristics specified in eq. 1
    of the `Feature Pyramid Network paper <https://arxiv.org/abs/1612.03144>`_.
    They keyword-only parameters ``canonical_scale`` and ``canonical_level``
    correspond respectively to ``224`` and ``k0=4`` in eq. 1, and
    have the following meaning: ``canonical_level`` is the target level of the pyramid from
    which to pool a region of interest with ``w x h = canonical_scale x canonical_scale``.

    Args:
        featmap_names (List[str]): the names of the feature maps that will be used
            for the pooling.
        output_size (List[Tuple[int, int]] or List[int]): output size for the pooled region
        sampling_ratio (int): sampling ratio for ROIAlign
        canonical_scale (int, optional): canonical_scale for LevelMapper
        canonical_level (int, optional): canonical_level for LevelMapper
    """

    # pylint: disable=too-many-locals
    def __init__(
        self,
        featmap_names: List[str],
        output_size: Union[int, Tuple[int], List[int]],
        sampling_ratio: int,
        *,
        canonical_scale: int = 224,
        canonical_level: int = 4,
    ):
        super().__init__()
        if isinstance(output_size, int):
            output_size = (output_size, output_size)  # type: ignore[assignment]
        self.featmap_names = featmap_names
        self.sampling_ratio = sampling_ratio
        self.output_size = tuple(output_size)  # type: ignore[arg-type]
        self.scales: Optional[List[float]] = None
        self.map_levels: Optional[LevelMapper] = None
        self.canonical_scale = canonical_scale
        self.canonical_level = canonical_level

    @staticmethod
    def convert_to_roi_format(boxes: List[Tensor]) -> Tensor:
        """Convert boxes to ROI format and transfer to target device"""
        concat_boxes = torch.cat(boxes, dim=0)
        device, dtype = concat_boxes.device, concat_boxes.dtype
        ids = torch.cat(
            [
                torch.full_like(
                    b[:, :1], i, dtype=dtype, layout=torch.strided, device=device
                )
                for i, b in enumerate(boxes)
            ],
            dim=0,
        )
        rois = torch.cat([ids, concat_boxes], dim=1)
        return rois

    @staticmethod
    def infer_scale(feature: Tensor, original_size: Iterable[int]) -> float:
        """Infer scale of feature from original input size"""
        # assumption: the scale is of the form 2 ** (-k), with k integer
        feature_size = feature.shape[-2:]
        possible_scales: List[float] = []
        for feat_side, orig_side in zip(feature_size, original_size):
            approx_scale = float(feat_side) / float(orig_side)
            scale = 2 ** float(torch.tensor(approx_scale).log2().round())
            possible_scales.append(scale)
        assert possible_scales[0] == possible_scales[1]
        return possible_scales[0]

    def setup_scales(
        self,
        features: List[Tensor],
        image_shapes: List[Tuple[int, int]],
    ) -> None:
        """Method to setup scales for roi_align() function"""
        assert len(image_shapes) != 0
        max_x = 0
        max_y = 0
        for shape in image_shapes:
            max_x = max(shape[0], max_x)
            max_y = max(shape[1], max_y)
        original_input_shape = (max_x, max_y)

        scales = [self.infer_scale(feat, original_input_shape) for feat in features]
        # get the levels in the feature map by leveraging the fact that the network always
        # downsamples by a factor of 2 at each level.
        lvl_min = -torch.log2(torch.tensor(scales[0], dtype=torch.float32)).item()
        lvl_max = -torch.log2(torch.tensor(scales[-1], dtype=torch.float32)).item()
        self.scales = scales  # type: ignore[assignment]
        self.map_levels = init_level_mapper(  # type: ignore[assignment]
            int(lvl_min),
            int(lvl_max),
            canonical_scale=self.canonical_scale,
            canonical_level=self.canonical_level,
        )

    def forward(
        self,
        feat_map_dict: Dict[str, Tensor],
        boxes: List[Tensor],
        image_shapes: List[Tuple[int, int]],
    ) -> Tensor:
        """
        Args:
            feat_map_dict (OrderedDict[Tensor]): feature maps for each level. They are assumed to
                have all the same number of channels, but they can have different sizes.
            boxes (List[Tensor[N, 4]]): boxes to be used to perform the pooling operation, in
                (x1, y1, x2, y2) format and in the image reference size, not the feature map
                reference. The coordinate must satisfy ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
            image_shapes (List[Tuple[height, width]]): the sizes of each image before they
                have been fed to a CNN to obtain feature maps. This allows us to infer the
                scale factor for each one of the levels to be pooled.
        Returns:
            result (Tensor)
        """
        x_filtered = []
        for featmap_name, feat_maps in feat_map_dict.items():
            if featmap_name in self.featmap_names:
                x_filtered.append(feat_maps)
        num_levels = len(x_filtered)
        rois = self.convert_to_roi_format(boxes)
        if self.scales is None:
            self.setup_scales(x_filtered, image_shapes)

        scales = self.scales
        assert scales is not None

        if num_levels == 1:
            return roi_align.roi_align(
                x_filtered[0],
                rois,
                output_size=self.output_size,
                spatial_scale=scales[0],
                sampling_ratio=self.sampling_ratio,
            )

        mapper = self.map_levels
        assert mapper is not None

        levels = mapper(boxes)

        num_rois = len(rois)
        num_channels = x_filtered[0].shape[1]

        dtype, device = x_filtered[0].dtype, x_filtered[0].device
        result = torch.zeros(
            (
                num_rois,
                num_channels,
            )
            + self.output_size,
            dtype=dtype,
            device=device,
        )

        for level, (per_level_feature, scale) in enumerate(zip(x_filtered, scales)):
            idx_in_level = torch.where(levels == level)[0]
            rois_per_level = rois[idx_in_level]

            result_idx_in_level = roi_align.roi_align(
                per_level_feature,
                rois_per_level,
                output_size=self.output_size,
                spatial_scale=scale,
                sampling_ratio=self.sampling_ratio,
            )

            # result and result_idx_in_level's dtypes are based on dtypes of different
            # elements in x_filtered.  x_filtered contains tensors output by different
            # layers.  When autocast is active, it may choose different dtypes for
            # different layers' outputs.  Therefore, we defensively match result's dtype
            # before copying elements from result_idx_in_level in the following op.
            # We need to cast manually (can't rely on autocast to cast for us) because
            # the op acts on result in-place, and autocast only affects out-of-place ops.
            result[idx_in_level] = result_idx_in_level.to(result.dtype)

        return result

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(featmap_names={self.featmap_names}, "
            f"output_size={self.output_size}, sampling_ratio={self.sampling_ratio})"
        )



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Implements the Generalized R-CNN framework
Modifications include:
- Removed code related to training and ground truth targets
- Removed torch scripting related code
"""

from typing import Tuple, List, Dict
from collections import OrderedDict
from torch import nn, Tensor


class GeneralizedRCNN(nn.Module):
    """
    Main class for Generalized R-CNN.

    Args:
        backbone (nn.Module):
        rpn (nn.Module):
        roi_heads (nn.Module): takes the features + the proposals from the RPN and computes
            detections / masks from it.
        transform (nn.Module): performs the data transformation from the inputs to feed into
            the model
    """

    def __init__(
        self,
        backbone: nn.Module,
        rpn: nn.Module,
        roi_heads: nn.Module,
        transform: nn.Module,
    ):
        super().__init__()
        self.transform = transform
        self.backbone = backbone
        self.rpn = rpn
        self.roi_heads = roi_heads
        # used only on torchscript mode
        self._has_warned = False

    def forward(self, images: List[Tensor]) -> List[Dict[str, Tensor]]:
        """
        Args:
            images (list[Tensor]): images to be processed

        Returns:
            result (list[BoxList] or dict[Tensor]): the output from the model.
                During testing, it returns list[BoxList] contains additional fields
                like `scores`, `labels` and `mask` (for Mask R-CNN models).
        """
        original_image_sizes: List[Tuple[int, int]] = []
        for img in images:
            val = img.shape[-2:]
            assert len(val) == 2
            original_image_sizes.append((val[0], val[1]))

        images = self.transform(images)

        features = self.backbone(images.tensors)
        if isinstance(features, Tensor):
            features = OrderedDict([("0", features)])
        proposals = self.rpn(images, features)
        detections = self.roi_heads(features, proposals, images.image_sizes)
        detections = self.transform.postprocess(  # type: ignore[operator]
            detections, images.image_sizes, original_image_sizes
        )

        return detections



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CLI functions for PeekingDuck.
"""

import click

from peekingduck import __version__
from peekingduck.commands.core import init, run, verify_install
from peekingduck.commands.create_node import create_node
from peekingduck.commands.nodes import nodes


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """PeekingDuck is a modular computer vision inference framework.
    Developed by Computer Vision Hub at AI Singapore.
    """


cli.add_command(create_node)
cli.add_command(init)
cli.add_command(nodes)
cli.add_command(run)
cli.add_command(verify_install)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PeekingDuck Viewer GUI Creation Code
"""

# Technotes:
#   using __future__ and TYPE_CHECKING works with pylint 2.10.x but fails with 2.7.x
#
# from __future__ import annotations
# from typing import TYPE_CHECKING
# if TYPE_CHECKING:
#     from peekingduck.player.viewer import Viewer
#
# Did not import Viewer from peekingduck.viewer.viewer due to pylint 2.7 complaining
# about circular import.

from pathlib import Path
import tkinter as tk
from tkinter import ttk
from peekingduck.viewer.viewer_utils import load_image
from peekingduck.viewer.single_column_view import SingleColumnPlayListView

LOGO: str = "PeekingDuckLogo.png"
MIN_HEIGHT: int = 768
MIN_WIDTH: int = 1024
WIN_HEIGHT: int = 960
WIN_WIDTH: int = 1280
BTN_WIDTH_SPAN = 1


def create_window(viewer) -> None:  # type: ignore
    """Create PeekingDuck Viewer window with the following structure:
        +-+-------------------------------+-+
        | |  Logo       Name              | |
        | +-------------------------------+ |
        | |                               | |
        | |                               | |
        | |     Video               Play  | |
        | |     image               list  | |
        | |                               | |
        | |                               | |
        | +-------------------------------+ |
        | |           Controls            | |
        | |           StatusBar           | |
        +-+-------------------------------+-+

    Args:
        viewer (Viewer): PeekingDuck Viewer object

            Required event handlers
                on_exit()
                on_keypress()
                on_resize()
                on_add_pipeline()
                on_delete_pipeline()
                on_play_pipeline()
    """
    root = tk.Tk()
    root.title("PeekingDuck Viewer")
    # bind event handlers
    root.wm_protocol("WM_DELETE_WINDOW", viewer.on_exit)
    root.bind("<Configure>", viewer.on_resize)
    root.bind("<Key>", viewer.on_keypress)
    root.geometry(f"{WIN_WIDTH}x{WIN_HEIGHT}")
    root.minsize(MIN_WIDTH, MIN_HEIGHT)
    root.update()  # force update before mainloop() to get correct size
    viewer.root = root  # save main window
    # Tk technotes: Need to create footer before body to ensure footer controls do not
    #               get covered when image is zoomed in
    create_side_margins(viewer)
    create_header(viewer)
    create_footer(viewer)
    create_body(viewer)
    viewer.btn_hide_show_playlist_press()  # first toggle to hide playlist


def create_header(viewer) -> None:  # type: ignore
    """Create header with PeekingDuck logo and pipeline name

    Args:
        viewer (Viewer): PeekingDuck Viewer object
    """
    header_frm = ttk.Frame(viewer.root, name="header_frm")
    header_frm.pack(side=tk.TOP, fill=tk.X)
    viewer.tk_header_frm = header_frm
    # row 0: logo (left)
    logo_path = Path(__file__).parent / LOGO
    viewer.img_logo = load_image(str(logo_path), resize_pct=0.10)
    logo = tk.Label(header_frm, image=viewer.img_logo, anchor=tk.W)
    logo.grid(row=0, column=0, sticky="nw")
    viewer.tk_logo = logo
    # row 1: viewer header text (center)
    lbl = tk.Label(header_frm, text="Viewer Header", font=("TkFixedFont 16"))
    lbl.grid(row=1, column=0, sticky="nsew")
    viewer.tk_header = lbl
    num_col, _ = header_frm.grid_size()  # config column sizes
    for i in range(num_col):
        header_frm.grid_columnconfigure(i, weight=1)


def create_footer(viewer) -> None:  # type: ignore
    """Create footer with controls and status bar

    Args:
        viewer (Viewer): PeekingDuck Viewer object
    """
    footer_frm = ttk.Frame(viewer.root, name="footer_frm")
    footer_frm.pack(side=tk.BOTTOM, fill=tk.X)
    viewer.tk_footer_frm = footer_frm
    # row 0: spacer
    lbl = tk.Label(footer_frm)  # row spacer
    lbl.grid(row=0, column=0)
    # row 1: controls
    ctrl_frm = ttk.Frame(footer_frm, name="ctrl_frm")
    ctrl_frm.grid(row=1, column=0, sticky="ew")
    _create_controls(viewer, ctrl_frm)
    # row 2: status bar
    lbl = tk.Label(footer_frm, anchor=tk.CENTER, text="Status bar text")
    lbl.grid(row=2, column=0, sticky="ew")
    viewer.tk_status_bar = lbl
    # row 3: spacer
    lbl = tk.Label(footer_frm)  # row spacer
    lbl.grid(row=3, column=0)

    num_col, _ = footer_frm.grid_size()  # config column sizes
    for i in range(num_col):
        footer_frm.grid_columnconfigure(i, weight=1)


def create_side_margins(viewer) -> None:  # type: ignore
    """Create left and right side margins

    Args:
        viewer (Viewer): PeekingDuck Viewer object
    """
    left_margin_frm = ttk.Frame(
        viewer.root, name="left_margin_frm", width=50, height=100
    )
    left_margin_frm.pack(side=tk.LEFT, fill=tk.NONE, expand=False)
    right_margin_frm = ttk.Frame(
        viewer.root, name="right_margin_frm", width=50, height=100
    )
    right_margin_frm.pack(side=tk.RIGHT, fill=tk.NONE, expand=False)


def create_body(viewer) -> None:  # type: ignore
    """Create body with video image and playlist in side panel

    Args:
        viewer (Viewer): PeekingDuck Viewer object
    """
    body_frm = ttk.Frame(viewer.root, name="body_frm")
    body_frm.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    viewer.tk_body_frm = body_frm
    #
    # right side panel for playlist
    #
    panel_frm = ttk.Frame(body_frm, name="panel_frm")
    panel_frm.pack(side=tk.RIGHT, fill=tk.Y)
    viewer.tk_panel_frm = panel_frm
    # playlist top/bottom spacers:
    # need these for video image to center properly after hiding playlist
    lbl = tk.Label(panel_frm)
    lbl.pack(side=tk.TOP)
    lbl = tk.Label(panel_frm)
    lbl.pack(side=tk.BOTTOM)
    # playlist
    playlist_frm = ttk.Frame(panel_frm, name="playlist_frm")
    playlist_frm.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    viewer.tk_playlist_frm = playlist_frm
    playlist_view = SingleColumnPlayListView(
        playlist=viewer.playlist, root=playlist_frm
    )
    viewer.tk_playlist_view = playlist_view
    playlist_view.register_callback("add", viewer.on_add_pipeline)
    playlist_view.register_callback("delete", viewer.on_delete_pipeline)
    playlist_view.register_callback("run", viewer.on_run_pipeline)
    viewer.playlist_show = True

    # video image
    image_frm = ttk.Frame(body_frm, name="image_frm")
    image_frm.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    viewer.tk_image_frm = image_frm
    output_image = tk.Label(image_frm)
    output_image.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
    viewer.tk_output_image = output_image


def _create_controls(viewer, ctrl_frm: ttk.Frame) -> None:  # type: ignore
    """Create 3 rows of controls: slider/progress bar/buttons

    Args:
        viewer (Viewer): PeekingDuck Viewer object
        body_frm (ttk.Frame): container frame for controls
    """
    #
    # row 0: info controls
    #
    # slider
    slider = ttk.Scale(
        ctrl_frm,
        orient=tk.HORIZONTAL,
        from_=1,
        to=100,
        command=viewer.sync_slider_to_frame,
    )
    slider.grid(row=0, column=0, columnspan=95, sticky="nsew")
    viewer.tk_scale = slider
    slider.grid_remove()  # hide it first
    # progress bar
    progress_bar = ttk.Progressbar(
        ctrl_frm,
        orient=tk.HORIZONTAL,
        length=100,
        mode="determinate",
        value=0,
        maximum=100,
    )
    progress_bar.grid(row=0, column=0, columnspan=95, sticky="nsew")
    viewer.tk_progress = progress_bar
    # frame number
    lbl = tk.Label(ctrl_frm, text="0", anchor=tk.W)
    lbl.grid(row=0, column=95, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_lbl_frame_num = lbl
    #
    # row 1: buttons
    #
    btn_play = ttk.Button(ctrl_frm, text="Play", command=viewer.btn_play_stop_press)
    btn_play.grid(row=1, column=0, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_btn_play = btn_play
    btn_zoom_out = ttk.Button(ctrl_frm, text="-", command=viewer.btn_zoom_out_press)
    btn_zoom_out.grid(row=1, column=92, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_btn_zoom_out = btn_zoom_out
    lbl = tk.Label(ctrl_frm, text="100%")
    lbl.grid(row=1, column=93, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_lbl_zoom = lbl
    btn_zoom_in = ttk.Button(ctrl_frm, text="+", command=viewer.btn_zoom_in_press)
    btn_zoom_in.grid(row=1, column=94, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_btn_zoom_in = btn_zoom_in
    # spacer: without this, GUI will resize and flicker when frame number is updated
    lbl = tk.Label(ctrl_frm, text="          ")
    lbl.grid(row=1, column=95, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    #
    # row 2: playlist button
    #
    btn_playlist = ttk.Button(
        ctrl_frm, text="Playlist", command=viewer.btn_hide_show_playlist_press
    )
    btn_playlist.grid(row=2, column=94, columnspan=BTN_WIDTH_SPAN, sticky="nsew")
    viewer.tk_btn_playlist = btn_playlist
    # spacer: without this, GUI will resize and flicker when frame number is updated
    lbl = tk.Label(ctrl_frm, text="          ")
    lbl.grid(row=2, column=95, columnspan=BTN_WIDTH_SPAN, sticky="nsew")

    num_col, _ = ctrl_frm.grid_size()  # config column sizes
    for i in range(num_col):
        ctrl_frm.grid_columnconfigure(i, weight=1)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorRT model for PeekingDuck"""

from typing import Any, List, Tuple
import numpy as np
import tensorrt as trt  # pylint: disable=import-error
import pycuda.driver as cuda  # pylint: disable=import-error

# NB: need below autoinit import to create CUDA context!
import pycuda.autoinit  # pylint: disable=import-error, unused-import


class HostDeviceMem:
    """Encapsulation for host CUDA device"""

    def __init__(self, host_mem: Any, device_mem: Any):
        self.host = host_mem
        self.device = device_mem

    def __str__(self) -> str:
        return f"Host:\n{self.host}\nDevice:\n{self.device}"

    def __repr__(self) -> str:
        return self.__str__()


class TrtModel:  # pylint: disable=too-many-instance-attributes
    """YoloX TensorRT model class to load model engine and perform inference"""

    def __init__(self, engine_path: str, max_batch_size: int = 1):
        self.dtype = np.float32  # TensorRT support float32, not 64
        self.engine_path = engine_path
        self.max_batch_size = max_batch_size
        self.engine = self.load_engine(self.engine_path)
        self.inputs, self.outputs, self.bindings, self.stream = self.allocate_buffers()
        self.context = self.engine.create_execution_context()

    @staticmethod
    def load_engine(engine_path: str) -> Any:
        """Load TensorRT model engine file

        Args:
            engine_path (str): engine file full path

        Returns:
            TensorRT engine
        """
        trt.init_libnvinfer_plugins(None, "")
        trt_runtime = trt.Runtime(trt.Logger(trt.Logger.WARNING))
        with open(engine_path, "rb") as trt_engine_file:
            engine_data = trt_engine_file.read()
        engine = trt_runtime.deserialize_cuda_engine(engine_data)
        return engine

    def allocate_buffers(self) -> Tuple[List[Any], List[Any], List[Any], Any]:
        """Allocate CUDA working memory buffers

        Returns:
            (Tuple[List[Any], List[Any], List[Any], Any]): List of input, output,
                                                           bindings, stream buffers
        """
        inputs = []
        outputs = []
        bindings = []
        stream = cuda.Stream()

        for binding in self.engine:
            size = (
                trt.volume(self.engine.get_binding_shape(binding)) * self.max_batch_size
            )
            host_mem = cuda.pagelocked_empty(size, self.dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            bindings.append(int(device_mem))

            if self.engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                outputs.append(HostDeviceMem(host_mem, device_mem))

        return inputs, outputs, bindings, stream

    def __call__(self, data: np.ndarray, batch_size: int = 1) -> np.ndarray:
        """To allow making inference calls via `model(img)`

        Args:
            data (np.ndarray): input image data
            batch_size (int): inference batch size. Default = 1.

        Returns:
            (np.ndarray): inference result
        """
        data = data.astype(self.dtype)
        np.copyto(self.inputs[0].host, data.ravel())

        for inp in self.inputs:
            cuda.memcpy_htod_async(inp.device, inp.host, self.stream)

        self.context.execute_async(
            batch_size=batch_size,
            bindings=self.bindings,
            stream_handle=self.stream.handle,
        )

        for out in self.outputs:
            cuda.memcpy_dtoh_async(out.host, out.device, self.stream)

        self.stream.synchronize()
        res = [out.host.reshape(batch_size, -1) for out in self.outputs][0]
        # reshape the linear (1, N) res into YoloX-friendly shape (1, M, 85)
        result = res.reshape(1, -1, 85)
        return result
